/***************************************************************************
 *   Copyright (C) 2006-2007 by Guda Anton                                 *
 *   atu near dmeti dot dp point ua                                        *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <mntent.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <dirent.h>

#include "config.h"

#include <dbus/dbus.h>
#include <libhal.h>
#include <libhal-storage.h>

#define UDI_BASE "/org/freedesktop/Hal/devices/"
#define HAL_BASE "org.freedesktop.Hal"
#define MP_HOME  "/media/"
#define DEV_MAX 256
#define MAX_OPTS 32
#define MAX_FST  64

struct haldev {
  char dev[PATH_MAX];      //* full path to device, like /dev/sda1
  char udi[PATH_MAX];      //* UDI path from/to HAL
  char mpoint[PATH_MAX];   //* mount point (if mounted)
  char label[PATH_MAX];    //* FS label 
  char product[PATH_MAX];  //* device name 
  char fstype[MAX_FST];    //* xfs, ext2, vfat...
  int  is_mounted;
  int  major, minor;       //* device numbers
};

void show_help(const char *pgm );
int  show_list();
int  do_mount( const char *target, const char *opts);
int  do_umount( const char *target );
int  is_real_dev( const char *fn, char *fullname );
int  is_link_dev( const char *fn, char *lnkname, char *devname );
int  is_in_fstab( const char *ent, int mp );
int  use_real_mount( const char *ent, const char *opts );
int  hal_mount( const struct haldev *hd, const char *mp );
// functions to work with device list
int fill_halinfo();
struct haldev* find_by_dev( const char *adev );
struct haldev* find_by_mount( const char *mp );
void   print_dev( const struct haldev *hd );
void   print_all_dev();
int csv2list( char *csv, char **list, int maxel );
void print_list( char **list, int maxel );
void add_defoptions( const struct haldev *hd );
int has_opts( const char *opt );

// verbose level
int verbose = 0;
// name/version
static const char *const pgm_name = PACKAGE " ver " VERSION " (C) Guda Anton 2007. Under GPLv2";

// global devices infos for hal. init by fill_halinfo
struct haldev* ghd[DEV_MAX];
// and current number of elements in it
int   n_ghd = 0;
// list of options to mount: -o ro,async,umask=027 -> { "ro", "async", "umask=027" }
char *lopts[MAX_OPTS];
// and numer of elements
int n_opts = 0;

// commands
#define CMD_SHOW   1
#define CMD_MOUNT  2
#define CMD_UMOUNT 3

int main(int argc, char *argv[])
{
  int op, rc = -1, i, cmd = -1;
  char *opts = 0, *xopts = 0;

  for( i=0; i<n_ghd; ++i ) // for save free
    ghd[n_ghd] = 0;
  for( i=0; i<MAX_OPTS; ++i )
    lopts[i] = 0;

  while( (op=getopt(argc,argv,"hvluo:") ) != -1  ) {
    switch(op) {
      case 'h': show_help( argv[0] ); return 0;
      case 'v': verbose++; break;
      case 'l': cmd = CMD_SHOW; break;
      case 'u': cmd = CMD_UMOUNT; break;
      case 'o': opts = optarg ; break;
      default: fprintf( stderr, "Unknown or bad option <%c>\n", op );
	       return 1;
    };
  };

  // copy of opts to parse
  if( opts )
    xopts = strdup( opts );
  n_opts = csv2list( xopts, lopts, MAX_OPTS );
  
  if( verbose ) {
    printf( "Opts: %d 0: \n", n_opts );
    print_list( lopts, MAX_OPTS );
  };

  if( cmd == -1 ) {   // nothing was explicit set: use show/mount
    if( optind >= argc ) {    // no args --show list
      cmd = CMD_SHOW;
    } else {
      cmd = CMD_MOUNT;
    };
  };

  switch( cmd ) {
    case CMD_SHOW:   rc = show_list(); break;
    case CMD_MOUNT:  rc = do_mount( argv[optind], opts ); break;
    case CMD_UMOUNT: rc = do_umount( argv[optind] ); break;
    default: fprintf( stderr, "%s: unknown command %d \n", argv[0], cmd );
	     rc = 1;
  };
      
  free( xopts );
  for( i=0; i<n_ghd; ++i )
    free( ghd[n_ghd] );
 
  if( verbose )
    printf("hmount: exit with %d\n", rc);
  return rc;
}

void show_help(const char *pgm )
{
  printf( "%s:\n mount / umount / list devices via HAL or mount(1) \n\n", pgm_name );
  printf( "Usage: \n hmount: \n%s [-o mount_opts] { device | major_minor | label | product } \n\n", pgm );
  printf( " umount:\n %s [-o umount_opts] -u { device | mountpoint } \n\n", pgm );
  printf( " list:\n %s [-o umount_opts] [-l] \n\n", pgm );
  printf( "General options: \n -h = help\n -v = increase verbose level\n" );
}

int show_list()
{
  int rc;
  if( verbose ) {
    printf( "%s: List of mountable/mounted devices\n", __FUNCTION__ );
  };
  if( (rc = fill_halinfo()) != 0 )
    return rc;
  print_all_dev();
  printf( "Found %d devices\n", n_ghd );
  return 0;
}

int do_mount( const char *target, const char *opts)
{
  char dev[PATH_MAX], lnk[PATH_MAX], mpx[PATH_MAX];
  struct haldev *hd;
  int rc;
  dev[0] = lnk[0] = 0;
  if( verbose ) {
    printf( "%s: will try to mount <%s> with opts <%s>\n",
	__FUNCTION__, target, opts );
  };

  // --------------------------------------- real dev --------------------
  if( is_real_dev( target, dev ) ) {
    if( verbose ) {
      printf( "%s: target <%s> is a real device <%s>\n", 
	  __FUNCTION__, target, dev );
    };
    if( is_in_fstab( dev, 0 ) ) 
      return use_real_mount( dev, opts );

  // ------------------------------------------link to real dev ----------   
  } else if( is_link_dev( target, lnk, dev ) ) {
    if( verbose ) {
      printf( "%s: target <%s> is a link <%s> to real device <%s>\n", 
	  __FUNCTION__, target, lnk, dev );
    };
    if( is_in_fstab( lnk, 0 ) ) 
      return use_real_mount( lnk, opts );

  // --------------------------------------- try mountpoint in fstab
  } else if( target[0] == '/' ) {
    if( is_in_fstab( target, 1 ) ) 
      return use_real_mount( target, opts );
    // TODO: not, may be a mountpoint from /media ?
    fprintf( stderr,  "%s: target <%s> starting with / must be real device\n"
	              "   or link to it\n"
		      "   or directory from /etc/fstab\n",
	     __FUNCTION__, target );
    return 3;
  };

  // ----------------------------------- HAL mount part ---------------------

  if( target[0] == ':' )  // skip ':' prefix in special names
    target++;

  if( (rc=fill_halinfo()) != 0 )
    return rc;

  // first try as device name
  if( target[0] != '/' ) {  //   partial name
    strcpy( dev, "/dev/" ); 
    strncat( dev, target, sizeof(dev)-1 );
    strncat( mpx, target, sizeof(mpx)-1 );
  } else {               // full name 
    const char *s = strrchr( target, '/' );
    if( s ) {
      ++s;
    } else {  // unlik
      s = "???";
    };
    printf( ".... s= <%s> \n", s );
    mpx[0] = 0; strncat( mpx, s, sizeof(mpx)-1 );
  };
  hd = find_by_dev( dev );
  if( hd ) {
    return hal_mount( hd, mpx );
  }

  // then try name based on major_minor scheme
  int d_major, d_minor, nf, i;
  nf = sscanf( target, "%d_%d", &d_major, &d_minor );
  if( nf == 2 ) {
    for( i=0; i < n_ghd; ++i ) {
      if( ghd[i]->major == d_major && ghd[i]->minor == d_minor ) {
	return hal_mount( ghd[i], target );
      };
    };
  };

  // then try label (first found)
  for( i=0; i < n_ghd; ++i ) {
    if( strcmp( ghd[i]->label, target ) == 0 ) {
      return hal_mount( ghd[i], target );
    };
  };

  // then try product (first found)
  for( i=0; i < n_ghd; ++i ) {
    if( strcmp( ghd[i]->product, target ) == 0 ) {
      return hal_mount( ghd[i], target );
    };
  };
  fprintf( stderr, "Fail to found device, corresponding target <%s>\n", target );

  return 1;
}

int  hal_mount( const struct haldev *hd, const char *mp )
{
  DBusMessage *msg = 0, *repl = 0;
  DBusConnection *dbus_conn;
  DBusError dberr;
  LibHalContext *hal_ctx;

  if( verbose ) {
    printf( "%s: try to HAL mount device <%s> on %s\n", 
            __FUNCTION__, hd->dev, mp );
    print_dev( hd );
  };
 
  // ---------------------------- real HAL mount ------------------------
  
  hal_ctx = libhal_ctx_new();
  if( !hal_ctx ) {
    fprintf( stderr, "%s: Fail to get libHAL context\n", __FUNCTION__ );
    return 3;
  };
	
  dbus_error_init (&dberr);
  if( verbose ) {
    printf( "%s: try to connect systen DBUS: \n", __FUNCTION__ );
  };
  
  dbus_conn = dbus_bus_get(DBUS_BUS_SYSTEM, &dberr);
  if (dbus_error_is_set( &dberr)) {
    fprintf( stderr, "%s: Fail to connect system DBUS: %s : %s\n",
	__FUNCTION__, dberr.name, dberr.message);
    dbus_error_free (&dberr);
    return 2;
  }
  
  libhal_ctx_set_dbus_connection( hal_ctx, dbus_conn );
  
  if ( !libhal_ctx_init (hal_ctx, &dberr) ) {
    fprintf( stderr, "%s Failed to init libhal context: <%s>: %s \n",
	     __FUNCTION__, dberr.name, dberr.message );
    dbus_error_free ( &dberr );
    libhal_ctx_free( hal_ctx );
    return 3;
  }
  
  if (!(msg = dbus_message_new_method_call (HAL_BASE, hd->udi,
                        HAL_BASE ".Device.Volume", "Mount"))) {
    fprintf( stderr, "Could not create dbus message for %s", hd->udi );
    libhal_ctx_free( hal_ctx );
    return 4;
  }
  const char *xfstype = (hd->fstype);
  add_defoptions( hd );
  char** xtestopts = lopts;
  if (!dbus_message_append_args (msg, 
        DBUS_TYPE_STRING, &mp, 
        DBUS_TYPE_STRING, &xfstype,
	DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &xtestopts, n_opts,
        DBUS_TYPE_INVALID)) {
    fprintf( stderr, "Could not append args to dbus message for %s", hd->udi);
    dbus_message_unref( msg );
    libhal_ctx_free( hal_ctx );
    return 5;
  }
  
  if (!(repl = dbus_connection_send_with_reply_and_block (dbus_conn, msg, 10000, &dberr)) ||
       dbus_error_is_set (&dberr)) {
    fprintf( stderr, "Mount failed for %s\n%s : %s\n", 
	     hd->dev, dberr.name, dberr.message);
    dbus_error_free (&dberr);
  };
  dbus_message_unref( msg );
  if( repl )
    dbus_message_unref( repl );

  libhal_ctx_free( hal_ctx );
  return 0;
}

int do_umount( const char *target )
{
  DBusMessage *msg = 0;
  DBusMessage *repl = 0;
  DBusConnection *dbus_conn;
  DBusError dberr;
  LibHalContext *hal_ctx;
  struct haldev *hd = 0;
  int rc;
  
  if( verbose ) {
    printf( "%s: will try to umount <%s> \n",
	__FUNCTION__, target );
  }

  if( ! target ) {
    fprintf( stderr, "%s: need target to umount \n", __FUNCTION__ );
    return 1;
  };

  if( (rc=fill_halinfo()) != 0 )
    return rc;

  if( ( hd = find_by_dev( target) ) == 0 ) {
    hd = find_by_mount( target );
  }

  if( ! hd ) {
    fprintf( stderr, "%s: fail to found target for umount: <%s>\n",
	     __FUNCTION__, target );
    return 5;
  }

  // find udi by target: dev, mp, ?label?
  
  hal_ctx = libhal_ctx_new();
  if( !hal_ctx ) {
    fprintf( stderr, "%s: Fail to get libHAL context\n", __FUNCTION__ );
    return 3;
  };
	
  dbus_error_init (&dberr);
  if( verbose ) {
    printf( "%s: try to connect systen DBUS: \n", __FUNCTION__ );
  };
  
  dbus_conn = dbus_bus_get(DBUS_BUS_SYSTEM, &dberr);
  if (dbus_error_is_set( &dberr)) {
    fprintf( stderr, "%s: Fail to connect system DBUS: %s : %s\n",
	__FUNCTION__, dberr.name, dberr.message);
    dbus_error_free (&dberr);
    return 2;
  }
  
  libhal_ctx_set_dbus_connection( hal_ctx, dbus_conn );
  
  if ( !libhal_ctx_init (hal_ctx, &dberr) ) {
    fprintf( stderr, "%s Failed to init libhal context: <%s>: %s \n",
	     __FUNCTION__, dberr.name, dberr.message );
    dbus_error_free ( &dberr );
    libhal_ctx_free( hal_ctx );
    return 3;
  }
  

  msg = dbus_message_new_method_call (HAL_BASE, hd->udi,
      HAL_BASE ".Device.Volume",
      "Unmount");
  if (msg == NULL) {
    fprintf( stderr, "Could not create dbus message for %s", hd->udi );
    libhal_ctx_free( hal_ctx );
    return 4;
  }
  
  char** xtestopts = lopts;
  if (!dbus_message_append_args (msg, 
	DBUS_TYPE_ARRAY, DBUS_TYPE_STRING, &xtestopts, n_opts,
        DBUS_TYPE_INVALID)) {
    fprintf( stderr, "Could not append args to dbus message for %s", hd->udi);
    dbus_message_unref( msg );
    libhal_ctx_free( hal_ctx );
    return 5;
  }
  
  if (!(repl = dbus_connection_send_with_reply_and_block (dbus_conn, msg, 10000, &dberr)) ||
       dbus_error_is_set (&dberr)) {
    fprintf( stderr, "Mount failed for %s\n%s : %s\n", 
	     hd->dev, dberr.name, dberr.message);
    dbus_error_free (&dberr);
  };
  dbus_message_unref( msg );
  if( repl )
    dbus_message_unref( repl );

  
  libhal_ctx_free( hal_ctx );
  return 0;
}

/* checks, if given file is _real_ block device */
int is_real_dev( const char *fn, char *fullname )
{
  char fname[PATH_MAX];
  const char *f;
  struct stat st;
  if( !fn || !fn[0] )
    return 0;
  if( fn[0] == '/' ) { // full path
    f = fn;
    if( strncmp( fn, "/dev/", 5 ) != 0 ) // not in /dev/
      return 0;
  } else {
    strcpy( fname, "/dev/" );
    strncat( fname, fn, sizeof(fname)-1 );
    f = fname;
  };
  if( lstat( f, &st ) != 0 )
    return 0;
  if( ! S_ISBLK( st.st_mode ) )
    return 0;
  if( fullname )
    strncpy( fullname, f, PATH_MAX );
  return 1;
}

/* checks, if given file is a link to real block device */
int is_link_dev( const char *fn, char *lnkname, char *devname )
{
  char fname[PATH_MAX];
  char dn[PATH_MAX], devx[PATH_MAX];
  const char *f;
  struct stat st, st_d;
  if( !fn || !fn[0] )
    return 0;
  if( fn[0] == '/' ) { // full path
    f = fn;
    if( strncmp( fn, "/dev/", 5 ) != 0 ) // not in /dev/
      return 0;
  } else {
    strcpy( fname, "/dev/" );
    strncat( fname, fn, sizeof(fname)-1 );
    f = fname;
  };
  if( lstat( f, &st ) != 0 || stat( f, &st_d) != 0 )
    return 0;
  if( ! S_ISLNK( st.st_mode ) || ! S_ISBLK( st_d.st_mode ) )
    return 0;
  if( readlink( f, dn, sizeof(dn) ) < 1 ) {
    fprintf( stderr, "%s: fail to read link <%s>: %s\n", 
	     __FUNCTION__, f, strerror(errno) );
    return 0;
  };
  if( is_real_dev( dn, devx ) ) { 
    if( lnkname )
      strncpy( lnkname, f, PATH_MAX );
    if( devname )
      strncpy( devname, devx, PATH_MAX );
    return 1;
  };
  return 0;
}

int is_in_fstab( const char *ent, int mp )
{
  FILE *fst;
  struct mntent *me;
  int rc = 0;
  if( verbose ) {
    printf( "%s: try to find <%s> in fstab as %s \n",
	    __FUNCTION__, ent, mp ? "mountpoint" : "device" );
  };
  fst = setmntent( "/etc/fstab", "r" );
  if( ! fst ) {
    fprintf( stderr, "%s: fail to open fstab: %s\n", 
	__FUNCTION__, strerror(errno) );
    return 0;
  };

  while( ( me = getmntent(fst) ) != 0 ) {
    if(mp) {
      if( strcmp( ent, me->mnt_dir ) == 0 ) {
	rc = 1;
	break;
      }
    } else {
      if( strcmp( ent, me->mnt_fsname ) == 0 ) {
	rc = 1;
	break;
      }
    };
  }

  endmntent( fst );
  if( verbose ) {
    printf( "%s: .. %s\n", __FUNCTION__, rc ? "Found" : "Not found" );
    if( rc )
    printf( "%s: ... fs: <%s> mp: <%s> type: <%s> opts: <%s> \n",
	     __FUNCTION__, me->mnt_fsname, me->mnt_dir,
	     me->mnt_type, me->mnt_opts );
  }
  return rc;
}
	
int use_real_mount( const char *ent, const char *opts )
{
  int child, wrc, status;
  if( verbose ) {
    printf( "%s: try to use real mount for <%s> with <%s>\n",
	    __FUNCTION__, ent, opts );
  };
  if( ( child=fork() ) == -1 ) {
    fprintf( stderr, "%s: fail to fork: %s\n", __FUNCTION__, strerror(errno) );
    return 7;
  };

  // -------------------- child part
  if( child == 0 ) {
    if( opts )
      execlp( "mount", "mount", ent, "-o", opts , NULL );
    else
      execlp( "mount", "mount", ent, NULL );
    fprintf( stderr, "%s: child: fail to exec mount: %s\n", __FUNCTION__, strerror(errno) );
    exit( 7 );
  };

  // ---------------- parent part
  if( wait( &status ) == -1 ) {
    fprintf( stderr, "%s: fail to wait: %s\n", __FUNCTION__, strerror(errno) );
    return 8;
  };
  if( WIFSIGNALED( status ) ) {
    fprintf( stderr, "%s: mount was killed by signal %d\n", 
	     __FUNCTION__, WTERMSIG(status) );
    return 10;
  };
  if( WIFEXITED( status ) ) {
    wrc = WEXITSTATUS( status );
    if( wrc != 0 ) {
      fprintf( stderr, "%s: mount exited with error %d\n", 
	       __FUNCTION__, wrc );
      return 10;
    }
    return 0;
  };
  fprintf( stderr, "%s: mount finished by unknown way %X\n", 
           __FUNCTION__, status );
  return 11;
}

// ------------------------------------- fill global ghl[0 .. n_ghl-1] -------
int fill_halinfo()
{
  DBusConnection *dbus_conn;
  DBusError dberr;
  LibHalContext *hal_ctx;
  char **dnames;
  char *devcat, *volusage, *dev, *mp, *fstype, *fslabel, *product;
  int ndev, i, im, ign, hotpl, hotpl_parent, dev_major, dev_minor;
  char* parent_udi;
  struct haldev *hd;

  n_ghd = 0;

  hal_ctx = libhal_ctx_new();
  if( !hal_ctx ) {
    fprintf( stderr, "%s: Fail to get libHAL context\n", __FUNCTION__ );
    return 3;
  };
	
  dbus_error_init (&dberr);
  if( verbose ) {
    printf( "%s: try to connect systen DBUS: \n", __FUNCTION__ );
  };
  
  dbus_conn = dbus_bus_get(DBUS_BUS_SYSTEM, &dberr);
  if (dbus_error_is_set( &dberr)) {
    fprintf( stderr, "%s: Fail to connect system DBUS: %s : %s\n",
	__FUNCTION__, dberr.name, dberr.message);
    dbus_error_free (&dberr);
    return 2;
  }
  
  libhal_ctx_set_dbus_connection( hal_ctx, dbus_conn );
  
  if ( !libhal_ctx_init (hal_ctx, &dberr) ) {
    fprintf( stderr, "%s Failed to init libhal context: <%s>: %s \n",
	     __FUNCTION__, dberr.name, dberr.message );
    dbus_error_free ( &dberr );
    libhal_ctx_free( hal_ctx );
    return 3;
  }

  dnames = libhal_get_all_devices( hal_ctx, &ndev, &dberr );
  if( !dnames ) {
    fprintf( stderr, "%s: fail to get devices list: %s %s\n",
	     __FUNCTION__, dberr.name, dberr.message );
    dbus_error_free ( &dberr );
    libhal_ctx_free( hal_ctx );
    return 3;
  };
  if( verbose > 1) {
    printf( "%s: got %d devices:\n", __FUNCTION__, ndev );
    for( i=0; i<ndev; ++i )
      printf( "  <%s>\n", dnames[i] );
  };
  for( i=0; i<ndev; ++i ) {
    if( n_ghd >= DEV_MAX -1 ) {
      fprintf( stderr, 
	  "%s: Warning: to many (%d) devices detected, some was skipped\n",
	  __FUNCTION__, n_ghd );
      break;
    };

    // --- skip non-volume devices 
    devcat = libhal_device_get_property_string( hal_ctx, dnames[i],  
	       "info.category", &dberr );
    if( dbus_error_is_set( &dberr ) )  {
      dbus_error_free( &dberr );
      continue;
    }
    if( !devcat || strcmp( devcat, "volume" ) != 0 ) {
      libhal_free_string( devcat ); devcat = 0;
      continue;
    }
    libhal_free_string( devcat ); devcat = 0;

    // -- skip volumes w/o filesystems
    volusage = libhal_device_get_property_string( hal_ctx, dnames[i],  
	       "volume.fsusage", &dberr );
    if( dbus_error_is_set( &dberr ) )  {
      dbus_error_free( &dberr );
      continue;
    }
    if( ! volusage || strcmp( volusage, "filesystem" ) != 0 ) {
      libhal_free_string( volusage ); volusage = 0;
      continue;
    }
    libhal_free_string( volusage ); volusage = 0;

    // TODO: check here: ( storage.hotpluggable == true )  for self or parent
    hotpl = libhal_device_get_property_bool( hal_ctx, dnames[i],  
	       "storage.hotpluggable", &dberr );
    if( dbus_error_is_set( &dberr ) )  {
      dbus_error_free( &dberr );
      // continue;
    }
    if( ! hotpl ) {   // parent may by hotpluggable

      parent_udi = libhal_device_get_property_string( hal_ctx, dnames[i],  
		 "info.parent", &dberr );   // find parent
      if( dbus_error_is_set( &dberr ) )  {
	dbus_error_free( &dberr );
	continue;
      }

      hotpl_parent = libhal_device_get_property_bool( hal_ctx, parent_udi,  
		 "storage.hotpluggable", &dberr );
      libhal_free_string( parent_udi ); parent_udi = 0;
      if( dbus_error_is_set( &dberr ) )  {
	dbus_error_free( &dberr );
	continue;
      }
      hotpl = hotpl_parent;
    };
    if( ! hotpl ) {  // not self, not parent
      continue;
    };
    
    // ignore ignored devices :-)
    ign = libhal_device_get_property_bool( hal_ctx, dnames[i],  
	       "volume.ignore", &dberr );
    if( dbus_error_is_set( &dberr ) )  {
      dbus_error_free( &dberr );
      continue;
    }
    if( ign ) 
      continue;

    // --- device file: /dev/sda1 ....
    dev = libhal_device_get_property_string( hal_ctx, dnames[i],  
	       "block.device", &dberr );
    if( dbus_error_is_set( &dberr ) )  {
      fprintf( stderr, "%s: fail to get block.device for %s: %s %s\n",
	     __FUNCTION__, dnames[i], dberr.name, dberr.message );
      dbus_error_free( &dberr );
      continue;
    }

    // --- mount point
    mp = libhal_device_get_property_string( hal_ctx, dnames[i],  
	       "volume.mount_point", &dberr );
    if( !mp ) 
      mp = "";
    if( dbus_error_is_set( &dberr ) ) 
      dbus_error_free( &dberr );
    
    // --- FS type: vfat, xfs .....
    fstype = libhal_device_get_property_string( hal_ctx, dnames[i],  
	       "volume.fstype", &dberr );
    if( !fstype ) 
      fstype = "";
    if( dbus_error_is_set( &dberr ) ) 
      dbus_error_free( &dberr );
    
    // --- FS label
    fslabel = libhal_device_get_property_string( hal_ctx, dnames[i],  
	       "volume.label", &dberr );
    if( !fslabel ) 
      fslabel = "";
    if( dbus_error_is_set( &dberr ) ) 
      dbus_error_free( &dberr );
    
    // --- device product
    product = libhal_device_get_property_string( hal_ctx, dnames[i],  
	       "info.product", &dberr );
    if( !product ) 
      product = "";
    if( dbus_error_is_set( &dberr ) ) 
      dbus_error_free( &dberr );

    // ---- IS mounted?
    im = libhal_device_get_property_bool( hal_ctx, dnames[i],  
	       "volume.is_mounted", &dberr );
    if( dbus_error_is_set( &dberr ) ) 
      dbus_error_free( &dberr );

    // device numbers 
    dev_major = libhal_device_get_property_int( hal_ctx, dnames[i],  
	       "block.major", &dberr );
    if( dbus_error_is_set( &dberr ) ) 
      dbus_error_free( &dberr );
    dev_minor = libhal_device_get_property_int( hal_ctx, dnames[i],  
	       "block.minor", &dberr );
    if( dbus_error_is_set( &dberr ) ) 
      dbus_error_free( &dberr );

    // --------- fill element ---------------------------------------
    hd = ghd[n_ghd] = (struct haldev*) malloc( sizeof( struct haldev ));
    bzero( hd, sizeof( struct haldev ) );
    if( ! hd ) {
      fprintf( stderr, "%s: Out of memory with %d elements\n", 
	       __FUNCTION__, n_ghd );
      break;
    };
    strncat( hd->dev, dev, sizeof(hd->dev) );
    strncat( hd->udi, dnames[i], sizeof(hd->dev) );
    strncat( hd->mpoint, mp, sizeof(hd->mpoint) );
    strncat( hd->label, fslabel, sizeof(hd->label) );
    strncat( hd->fstype, fstype, sizeof(hd->fstype) );
    strncat( hd->product, product, sizeof(hd->product) );
    hd->is_mounted = im;
    hd->major = dev_major;
    hd->minor = dev_minor;
    n_ghd++;
    libhal_free_string( dev );
    if( mp[0] ) 
      libhal_free_string( mp );
    if( fstype[0] ) 
      libhal_free_string( fstype );
    if( fslabel[0] ) 
      libhal_free_string( fslabel );
    if( product[0] ) 
      libhal_free_string( product );
  };

  libhal_free_string_array( dnames );
  libhal_ctx_free( hal_ctx );
  if( verbose ) 
    printf( "%s: found %d devices\n", __FUNCTION__, n_ghd );
  return 0;
}


struct haldev* find_by_dev( const char *adev )
{
  int i;
  for( i=0; i < n_ghd; i++ ) {
    if( strcmp( adev, ghd[i]->dev ) == 0 )
      return ghd[i];
  }
  return 0;
}

struct haldev* find_by_mount( const char *mp )
{
  int i, l;
  char fmp[PATH_MAX]; // full path to mount point

  if( !mp || !mp[0] )
    return 0;

  fmp[0] = 0;
  if( mp[0] == '/' ) {                  // absolute path
    strncat( fmp, mp, sizeof(fmp)-1 );
  } else {                              // relative path: prepend /media/
    strcat( fmp, MP_HOME );
    strncat( fmp, mp, sizeof(fmp)-1 );
  };
  l = strlen( fmp );
  if( l < 2 )
    return 0;
  if( fmp[l-1]  == '/' ) {            // trim right '/'
    fmp[l-1] = 0;
    --l;
  };

  for( i=0; i < n_ghd; i++ ) {
    if( strcmp( fmp, ghd[i]->mpoint ) == 0 )
      return ghd[i];
  }
  return 0;
}


//* outputs information about given device to stdout
//  
void   print_dev( const struct haldev *hd )
{
  printf( "<%s>=%d_%d ", hd->dev, hd->major, hd->minor );
  if( hd->is_mounted )
    printf( " at \"%s\" ", hd->mpoint );
  printf( "label: \"%s\" product: \"%s\" FS: \"%s\" \n", 
      hd->label, hd->product, hd->fstype );
  if( verbose > 0 )
    printf( "   udi: \"%s\"\n", hd->udi );
}

//* iterates all elements in ghd and prints info via print_dev
void   print_all_dev()
{
  int i;
  for( i=0; i < n_ghd; ++i )
    print_dev( ghd[i] );
}

int csv2list( char *csv, char **list, int maxel )
{
  int n;
  char *fc;
  if( csv == 0 || list == 0 || maxel < 2 )
    return 0;
  n = 0;
  fc = strtok( csv, "," );
  while( fc != 0 && n < maxel ) {
    printf( "%d: <%s>, ", n, fc );
    list[n] = fc;
    fc = strtok( 0, "," );
    ++n;
  };
  return n;
}

void print_list( char **list, int maxel )
{
  char *el;
  int i = 0;
  if( list == 0 || maxel < 1 )
    return;
  printf( "{ " );
  for( el = list[0]; el != 0 && i < maxel ; el=list[++i] ) {
    if( i != 0 )
      printf( ", " );
    printf( "%s", el );
  };
  printf( " }\n" );
}
  
// TODO: use simple config files
void add_defoptions( const struct haldev *hd )
{
  uid_t uid;
  static char uid_str[128];
  if( n_opts >= MAX_OPTS - 4 )
    return;
  if( strcmp( "vfat", hd->fstype ) == 0 ) { // options for vfat: uid=user ? showexec
    if( ! has_opts( "uid=" ) ) {
      uid = getuid();
      snprintf( uid_str, sizeof(uid_str)-2, "uid=%d", uid );
      lopts[n_opts++] = uid_str;
    };
    // TODO: after hal patch: add showexec
  }
}

int has_opts( const char *opt )
{
  int i;
  for( i=0; i< n_opts; ++i ) {
    if( strncmp( opt, lopts[i], strlen(opt) ) == 0 )
      return 1;
  };
  return 0;
}
